#ifndef __BOOST__STATISTICS
#define __BOOST__STATISTICS
#  include <boost/config.hpp> // for BOOST_NO_TEMPLATE_PARTIAL_SPECIALIZATION
#  include <complex>
#  include <cmath>

   /*
      based on --> Stat.H,v 1.1 2003/03/17 13:51:03 nbecker Exp
      by Neal D. Becker
   */

#  if defined(BOOST_MSVC)
      namespace std // MSVC library does not overload std::abs for base types
      {             // NOTE: Is this fixed in 7.1?
         inline float abs( float n )
         {
            return( ::fabsf( n ));
         }
         inline double abs( double n )
         {
            return( ::fabs( n ));
         }
         inline long double abs( long double n )
         {
            return( ::fabsl( n ));
         }
      }

#     pragma warning( push )
#     pragma warning( disable : 4227 ) // anachronism used : qualifiers on reference are ignored
#  endif

   namespace boost
   {
      template< typename T >
      struct stat_t
      {
         public:
            typedef T                                      type;
         public:
            static inline type                   convert( const T & const value )
            {
               return( value );
            }
      };

#     if !defined(BOOST_NO_TEMPLATE_PARTIAL_SPECIALIZATION)
         template< typename T >
         struct stat_t< std::complex< T > >
         {
            public:
               typedef typename std::complex<T>::value_type type;
            public:
               static inline type                convert( const T & const value )
               {
                  // NOTE: Is this what we want to do?
                  // should this be: return( std::abs( value ));????
                  return( value.real()); // [review]: extract the real component
               }
         };
#     endif

      template< typename T, typename CountT = int >
      class statistics
      {
         public:
            typedef T                                      value_type;
            typedef typename boost::stat_t< T >::type      return_type;
            typedef CountT                                 count_type;
         private: // internal state representation:
            count_type                 __count;
            value_type                 __sumX;
            return_type                __sumXsqr;
         public: // data insertion
            inline statistics &                  reset() throw();
            inline statistics &                  compute( const value_type & const ) throw();
            template< typename ForwardIterator >
            inline statistics &                  compute( ForwardIterator first, ForwardIterator last ) throw()
            {
               for( ; first != last; ++first )
                  compute( *first );
            }
         public:
            inline statistics &   operator+=( const T & const ) throw();
            inline statistics &   operator()( const T & const ) throw();
         public: // iterator support
            class iterator: public std::iterator< std::output_iterator_tag, void, void, void, void >
            {
               friend statistics;
               private:
                  statistics *   stat;
               public:
                  inline iterator & operator*()
                  {
                     return( *this );
                  }
                  inline iterator & operator++()
                  {
                     return( *this );
                  }
                  inline iterator & operator++( int )
                  {
                     return( *this );
                  }
               public:
                  inline iterator & operator=( const statistics::value_type & value ) throw()
                  {
                     if( stat )        stat -> compute( value );
                     return( *this );
                  }
               protected:
                  inline     iterator( const statistics * s ) throw(): stat( const_cast< statistics * >( s ))
                  {
                  }
               public:
                  inline     iterator( const iterator & i ) throw(): stat( i.stat )
                  {
                  }
                  inline     iterator() throw(): stat( 0 )
                  {
                  }
            };
         public:
            inline iterator                      begin() throw();
         public: // container interface support
            inline void                          push_front(       const T & const ) throw();
            inline void                          push_back(        const T & const ) throw();
            inline iterator                      insert( iterator, const T & const ) throw();
         public: // statistics
            inline value_type                    sumX()  const throw();
            inline value_type                    sumX2() const throw();
            inline count_type                    count() const throw();
         public:
            inline return_type                   xbar()  const throw();
            inline return_type                   x2bar() const throw();
         public:
            inline return_type                   mean()   const throw();
            inline return_type                   stddev() const throw();
            inline return_type                   var()    const throw();
         public:
            inline           statistics() throw();
      };
   }

   template< typename T, typename CountT >
   boost::statistics< T, CountT > & boost::statistics< T, CountT >::reset() throw()
   {
      __count   = 0;
      __sumX    = 0;
      __sumXsqr = 0;
      return( *this );
   }
   template< typename T, typename CountT >
   boost::statistics< T, CountT > & boost::statistics< T, CountT >::compute( const T & const value ) throw()
   {
      ++__count;
      __sumX    += value;
      __sumXsqr += ( value * value );
      return( *this );
   }
   template< typename T, typename CountT >
   boost::statistics< T, CountT > & boost::statistics< T, CountT >::operator+=( const T & const value ) throw()
   {
      return( compute( value ));
   }
   template< typename T, typename CountT >
   boost::statistics< T, CountT > & boost::statistics< T, CountT >::operator()( const T & const value ) throw()
   {
      return( compute( value ));
   }

   template< typename T, typename CountT >
   boost::statistics< T, CountT >::iterator boost::statistics< T, CountT >::begin() throw()
   {
      return( iterator( this ));
   }

   template< typename T, typename CountT >
   void boost::statistics< T, CountT >::push_front( const T & const value ) throw()
   {
      compute( value );
   }
   template< typename T, typename CountT >
   void boost::statistics< T, CountT >::push_back( const T & const value ) throw()
   {
      compute( value );
   }
   template< typename T, typename CountT >
   boost::statistics< T, CountT >::iterator boost::statistics< T, CountT >::insert
   (
      iterator pos,
      const T & const value
   ) throw()
   {
      compute( value );
      return( pos );
   }

   template< typename T, typename CountT >
   boost::statistics< T, CountT >::value_type boost::statistics< T, CountT >::sumX() const throw()
   {
      return( __sumX );
   }
   template< typename T, typename CountT >
   boost::statistics< T, CountT >::value_type boost::statistics< T, CountT >::sumX2() const throw()
   {
      return( __sumXsqr );
   }
   template< typename T, typename CountT >
   boost::statistics< T, CountT >::count_type boost::statistics< T, CountT >::count() const throw()
   {
      return( __count );
   }

   template< typename T, typename CountT >
   boost::statistics< T, CountT >::return_type boost::statistics< T, CountT >::xbar() const throw()
   {
      if( __count > 0 )                return( stat_t< T >::convert( __sumX / return_type( __count )));
      return( 0.0 );
   }
   template< typename T, typename CountT >
   boost::statistics< T, CountT >::return_type boost::statistics< T, CountT >::x2bar() const throw()
   {
      if( __count > 0 )                return( __sumXsqr / return_type( __count ));
      return( 0.0 );
   }

   template< typename T, typename CountT >
   boost::statistics< T, CountT >::return_type boost::statistics< T, CountT >::mean() const throw()
   {
      return( xbar());
   }
   template< typename T, typename CountT >
   boost::statistics< T, CountT >::return_type boost::statistics< T, CountT >::stddev() const throw()
   {
      return( std::sqrt( var()));
   }
   template< typename T, typename CountT >
   boost::statistics< T, CountT >::return_type boost::statistics< T, CountT >::var() const throw()
   {
      if( __count > 1 )
         return
         (
            stat_t< T >::convert
            (
               ( __sumXsqr - ( std::abs( __sumX * __sumX ) / return_type( __count ))) /
               return_type( __count - 1 )
            )
         );
      return( 0 );
   }

   template< typename T, typename CountT >
   boost::statistics< T, CountT >::statistics() throw():
      __count(   0 ),
      __sumX(    0 ),
      __sumXsqr( 0 )
   {
   }

#  if defined(BOOST_MSVC)
#     pragma warning( pop )
#  endif
#endif
